/***************************************************************************

Palm Chess Client
Copyright (C) 1999,2000 David Barr (davidbarr@iname.com)

Version 1.1 by Marco Brenco (joker@joker-games.com)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

****************************************************************************/

#include <PalmOS.h>
#include "chesscliRsc.h"

/* Some definitions belonging to Palm OS 4 and 5 needed here */
#define nativeFormat 2
#define vchrThumbWheelUp	0x012E		// optional thumb-wheel up
#define vchrThumbWheelDown	0x012F		// optional thumb-wheel down
#define vchrRockerUp		0x0132		// 5-way rocker up
#define vchrRockerDown		0x0133		// 5-way rocker down


/* Constants */
#define ALPHANUMERIC(c)	(((c)>='0' && (c)<='9') || ((c)>='a' && (c)<='z') || ((c)>='A' && (c)<='Z') || (c)=='_')
#define MYUPDATECODE (~frmRedrawUpdateCode)	// application specific update code
#define CREATORID  'CHCL'	// creator Id of this application
#define PREFSVERSION 1		// version of application preferences
#define BOARDTOP   0		// where on the screen to draw the Board
#define MESSAGETOP 16		// where on the screen to draw the top message
				//  NB: MESSAGEBOTTOM is defined in chesscliRsc.h
#define FONTHEIGHT 6		// the height of the font used for messages
#define FONTWIDTH 4		// the width of the font used for messages
#define NROWS 100		// number of rows of messages kept in buffer
#define NCHARS 40		// number of characters displayed per row
#define SCROLLAMOUNT 2		// number of rows scrolled with scroll buttons
#define BLANKS "                                        "	// as many blanks as NCHARS

static char PieceLabels[] = "-PNBRQKpnbrqk";	// String used to decode style 12


/* Global fields */
WinHandle OffImage = NULL;
UInt16 AutoOffTime = 0;		// it saves the auto-off time-out value when this app is started
int SquareSize = 16;		// the size of the board squares in pixels
char CurrentLine[256];		// the last line of text read from the server
char InitStr[256];		// initialization commands to be sent to the server
Boolean SentInitStr = false;	// whether InitStr has already been sent
UInt16 NetLibRefnum;
NetSocketRef Sock = -1;		// the socket number of our connection to the server
Boolean SockConnected = false;	// true after the socket is connected
BitmapPtr BitmapP[7*2*2];
UInt16 MainFormId = -1;		// ID of the last opened form (it only registers non-modal forms)

/* Fields used for displaying rows of the message form */
Boolean Current = true;		// whether the current buffer is displayed
char CurrentRows[NROWS][NCHARS+1];	// rows of the current buffer; index 0 is the bottom line
int CurrentNRows = 0;		// number of rows filled in the current buffer
char SavedRows[NROWS][NCHARS+1];	// rows of the saved buffer; index 0 is the bottom line
int SavedNRows = 0;		// number of rows filled in the saved buffer
int SavedBottomRow = 0;		// index into SavedRows of the bottom row when the saved buffer is displayed

/* Fields used during a game (played, observed or examined) */
char CurrentBoard[256];		// the last CurrentLine representing a board (begins with <12>)
char Board[8][8];		// the current Board, square by square
int MvTimeWh = 0;		// last clock time that white moved
int MvTimeBl = 0;		// last clock time that black moved
Boolean BlToMv = false;		// whether it is black's turn to move
UInt32 MvTimeLast = -1;		// last system time that either side moved
UInt32 MvTickLast = -1;		// last "tick" number when either side moved
int Relation = -3;		// player's relation to the current game, as defined in style 12:
				//  -3 isolated position (in this code = no involvement in a game)
				//  -2 we are observing a game being examined
				//   2 we are the examiner of the game
				//  -1 we are playing, it is our opponent's move
				//   1 we are playing and it is our move
				//   0 we are observing a game being played
int GameNumber = 0;		// the ID of the game being played, as attributed by the server
char *WhiteName = "";		// name of the white player
char *BlackName = "";		// name of the black player
Boolean Flip = false;		// true if the board must be flipped (White on top)

/* Fields used for board markers and blinks */
int NextMoveStartC = -1;	// first square tapped (column)
int NextMoveStartR = -1;	// first square tapped (row)
int NextMoveDestC  = -1;	// second square tapped (column)
int NextMoveDestR  = -1;	// second square tapped (row)
int LastMoveC = -1;		// square of last move (column)
int LastMoveR = -1;		// square of last move (row)
int LastMoveBlink = 0;		// 1 when last move's square has to blink

/* Preferences */
struct ChessPrefs_s {
  char host[80];
  char port[8];
  char user[80];
  char pass[80];
};
typedef struct ChessPrefs_s ChessPrefs_t;
ChessPrefs_t ChessPrefs = {"", "", "", ""};




/*
 * Return the ID of the currently active form, or -1 if it is not
 * safe to draw on the active form. This includes cases when the form
 * is covered by a window, e.g. when a menu is displayed on top.
 */
int
GetActiveFormId() {
  FormPtr frm;

  frm = FrmGetActiveForm();

  if (FrmGetWindowHandle(frm) == WinGetDrawWindow()) {
    return (int)FrmGetFormId(frm);
  } else {
    return -1;
  }
}



/*
 * Pops up a form immediately on top of the current one.
 * Unlike FrmPopupForm this is done without queueing events, therefore the
 * new form is displayed and becomes active before this function returns.
 * You can still use FrmReturnToForm() to close a form opened by PopupNowForm().
 */
void
PopupNowForm(UInt16 formId) {
  FormPtr frm;

  frm = FrmInitForm(formId);
  FrmSetActiveForm(frm);
  FrmDrawForm(frm);
}



/*
 * Pops up an error alert that displays an error number in hexadecimal format
 * in place of the ^1 variable of the message string.
 * Returns only when the user presses a button on the alert dialog.
 * The number of the button being pressed is not returned.
 */
void
PopupErrorAlert(UInt16 alertId, int error) {
  static char errhex[18];

  StrPrintF(errhex, "0x%x", error);
  (void) FrmCustomAlert(alertId, errhex, " ", " ");
}



/*
 * Add a new line to the current rows in the message form
 */
void
PutStr(char *s) {
  int top, height, len, i;
  RectangleType r, v;
  int formId;

  // Scroll rows in the current buffer
  CurrentNRows ++;
  if (CurrentNRows > NROWS) {
    CurrentNRows = NROWS;
  }

  for(i = CurrentNRows-1; i > 0; i --) {
    StrCopy(CurrentRows[i], CurrentRows[i-1]);
  }

  len = StrLen(s);

  // Copy the new row
  if(len < NCHARS) {
    StrCopy(CurrentRows[0], s);
    StrNCopy(CurrentRows[0]+len, BLANKS, NCHARS-len);	// fill with blanks
  } else {
    StrNCopy(CurrentRows[0], s, NCHARS);
  }
  CurrentRows[0][NCHARS] = '\0';

  formId = GetActiveFormId();

  // Modify the screen only if we are in the board form, or we are in
  // the message form and we are showing the current buffer
  if (formId == BoardForm || (formId == MessagesForm && Current)) {
    // Calculate the area to be scrolled
    if(formId == BoardForm) {
      top = BOARDTOP + SquareSize*8;
    } else {
      top = MESSAGETOP;
    }

    height = MESSAGEBOTTOM - top;

    r.topLeft.x = 0;
    r.topLeft.y = top;
    r.extent.x = 160;
    r.extent.y = height;

    // Scroll rows on screen and add a new row at the bottom
    WinScrollRectangle(&r, winUp, FONTHEIGHT, &v);
    FntSetFont(TermFontID);
    WinDrawChars(CurrentRows[0], NCHARS, 0, MESSAGEBOTTOM - FONTHEIGHT);
  }

  // Repeat for an eventual remainder of the string
  if(len > NCHARS) {
    PutStr(s+NCHARS);
  }
}



/*
 * Open the network library.
 * Return 0 if everything worked, or an error among:
 * sysErrLibNotFound, netErrOutOfMemory, netErrNoInterfaces, netErrPrefNotFound.
 */
Err
InitNetwork(void) {
  UInt16 err, ifErrs;

  if ((err = (UInt16) SysLibFind("Net.lib", &NetLibRefnum))) {
    return err;
  }

  err = (UInt16) NetLibOpen(NetLibRefnum, &ifErrs);
  if (err != 0 && err != netErrAlreadyOpen) {
    NetLibClose(NetLibRefnum, false);
    return err;
  }

  return 0;
}



/*
 * Read the application preferences from the preferences database
 */
void
LoadPrefs() {
  Int16 version;
  UInt16 prefsSize;

  prefsSize = sizeof(ChessPrefs_t);
  version = PrefGetAppPreferences(CREATORID, 1, &ChessPrefs, &prefsSize, false);

  if (version == noPreferenceFound || version != PREFSVERSION
   || prefsSize != sizeof(ChessPrefs_t)) {
    // Not the preferences saved by this version of the application.
    // Store some defaults in place of the (garbage) data that was read.
    ChessPrefs.host[0] = ChessPrefs.port[0] = ChessPrefs.user[0] =
     ChessPrefs.pass[0] = '\0';
  }
}



/*
 * Save game status information.
 */
void
SavePrefs() {
  PrefSetAppPreferences(CREATORID, 1, PREFSVERSION, &ChessPrefs,
   sizeof(ChessPrefs_t), false);
}



/*
 * Draw a square with its (eventual) piece.
 */
void
DrawSquare(int row, int col) {
  RectangleType r;
  int k, labellen;
  int x, y;
  char piece;
  UInt32 elapsed;
  int blink = 0;

  if (GetActiveFormId() != BoardForm) {
    return;
  }

  if (LastMoveBlink && LastMoveR == row && LastMoveC == col) {
    elapsed = TimGetTicks() - MvTickLast;

    if (elapsed >= SysTicksPerSecond()/3) {
      /* Restore original square color after 1/3 sec */
      LastMoveBlink = 0;
    } else {
      blink = 1;
    }
  }

  piece = Board[row][col];

  labellen = StrLen(PieceLabels);

  for(k = 0; k < labellen; k++) {
    if(piece == PieceLabels[k]) {
      break;
    }
  } /* end for k */

  r.topLeft.x = 0;
  r.topLeft.y = 0;

  if(k >= 7) { /* black pieces */
    r.topLeft.y += (SquareSize*2);
    k -= 6;
  }

  r.topLeft.y += (((row+col+k+blink)%2)*SquareSize);
  r.topLeft.x = k*SquareSize;
  r.extent.x = SquareSize;
  r.extent.y = SquareSize;

  x = (Flip ? 7-col : col) * SquareSize;
  y = (Flip ? row : 7-row) * SquareSize + BOARDTOP;

  WinCopyRectangle(OffImage, WinGetActiveWindow(), &r, x, y, winPaint);

  if ((row == NextMoveStartR && col == NextMoveStartC)
      || (row == NextMoveDestR && col == NextMoveDestC)) {
    /* Draw a marker */
    int size = SquareSize / 5;

    WinInvertLine(x+1, y+1, x+1+size, y+1);
    WinInvertLine(x+1, y+1, x+1, y+1+size);
    WinInvertLine(x+SquareSize-2, y+1, x+SquareSize-2-size, y+1);
    WinInvertLine(x+SquareSize-2, y+1, x+SquareSize-2, y+1+size);
    WinInvertLine(x+SquareSize-2, y+SquareSize-2, x+SquareSize-2-size, y+SquareSize-2);
    WinInvertLine(x+SquareSize-2, y+SquareSize-2, x+SquareSize-2, y+SquareSize-2-size);
    WinInvertLine(x+1, y+SquareSize-2, x+1+size, y+SquareSize-2);
    WinInvertLine(x+1, y+SquareSize-2, x+1, y+SquareSize-2-size);
  }
} /* end DrawSquare */



/*
 * Handle the case when a tap occurred over the board
 */
void
HandlePenOnBoard(int x, int y) {
  int r, c, prevStartR, prevStartC, prevDestR, prevDestC;
  char move[32];
  int piece;
  Err err;

  c = x/SquareSize;
  r = 7-(y-BOARDTOP)/SquareSize;
  if(Flip) {
    c = 7-c;
    r = 7-r;
  }

  /* Save old values of markers */
  prevStartR = NextMoveStartR;
  prevStartC = NextMoveStartC;
  prevDestR = NextMoveDestR;
  prevDestC = NextMoveDestC;

  if(c>=0 && c<=7 && r>=0 && r<=7) {
    if(NextMoveStartR == -1 || NextMoveDestR != -1) {
      /* Pen down on first square */
      NextMoveStartR = r;
      NextMoveStartC = c;
      NextMoveDestR = NextMoveDestC = -1;
    } else if (r == NextMoveStartR && c == NextMoveStartC) {
      /* First square = second square */
      NextMoveStartR = NextMoveStartC = -1;
      NextMoveDestR = NextMoveDestC = -1;
    } else {
      /* Pen down on second square */
      NextMoveDestR = r;
      NextMoveDestC = c;

      StrPrintF(move, "%c%d-%c%d", "abcdefgh"[NextMoveStartC], NextMoveStartR+1,
	        "abcdefgh"[NextMoveDestC], NextMoveDestR+1);

      // Look whether we are promoting a pawn
      piece = Board[NextMoveStartR][NextMoveStartC];

      if ((piece == 'P' && NextMoveDestR == 7)
        ||(piece == 'p' && NextMoveDestR == 0)) {
        // Display an alert to choose the piece
        switch(FrmAlert(PromotionAlert)) {
        case 0:	// queen
          StrCat(move, "=Q");
          break;
        case 1:	// rook
          StrCat(move, "=R");
          break;
        case 2:	// bishop
          StrCat(move, "=B");
          break;
        case 3:	// knight
          StrCat(move, "=N");
          break;
        default:
          break;
        }
      }

      if (SockConnected) {
        StrCat(move, "\n");
        NetLibSend(NetLibRefnum, Sock, (void *)move, StrLen(move),
         0, NULL, 0, SysTicksPerSecond()*5, &err);
      }
    }
  } else {
    /* Pen outside the Board */
    NextMoveStartR = NextMoveStartC = -1;
    NextMoveDestR = NextMoveDestC = -1;
  }

  if (prevStartR != -1 && prevStartR != NextMoveStartR) {
    DrawSquare(prevStartR, prevStartC);
  }

  if (NextMoveStartR != -1) {
    DrawSquare(NextMoveStartR, NextMoveStartC);
  }

  if (prevDestR != -1) {
    DrawSquare(prevDestR, prevDestC);
  }

  if (NextMoveDestR != -1) {
    DrawSquare(NextMoveDestR, NextMoveDestC);
  }
}



/*
 * Handle the case when a tap occurred over a message area
 */
void
HandlePenOnMessage(int x, int y) {
  int top, i, j, k, max;
  char (*buffer)[NCHARS+1];
  static char line[256];
  static char string[256];
  int separator1, separator2, quotes1, quotes2;
  char c;
  FormPtr frm;
  UInt16 index;
  FieldPtr fld;

  frm = FrmGetActiveForm();

  if (GetActiveFormId() == BoardForm) {
    top = BOARDTOP + SquareSize*8;
    index = FrmGetObjectIndex(frm, BoardFld);
  } else {
    top = MESSAGETOP;
    index = FrmGetObjectIndex(frm, MessagesFld);
  }

  fld = FrmGetObjectPtr(frm, index);	// pointer to the input field

  j = (MESSAGEBOTTOM-1-y) / FONTHEIGHT;
  i = x/FONTWIDTH;

  if (GetActiveFormId() == BoardForm || Current) {
    buffer = CurrentRows;
    max = CurrentNRows;
  } else {
    j += SavedBottomRow;
    buffer = SavedRows;
    max = SavedNRows;
  }

  if (i < 0 || i >= NCHARS || j < 0 || j >= max) {
    return;
  }

  // Create a string composed by the tapped row and the two rows above and below
  if (j+1 < max) {
    i += NCHARS;
    StrCopy(line, buffer[j+1]);
    StrCat(line, buffer[j]);
  } else {
    StrCopy(line, buffer[j]);
  }

  if (j-1 >= 0) {
    StrCat(line, buffer[j-1]);
  }

  quotes1 = quotes2 = -1;
  separator1 = -1;
  separator2 = StrLen(line);

  // Now scan the string up to the i-th char looking for the last
  // non alphanumeric char and the last quotes char
  for (k = 0; k < i; k ++) {
    c = line[k];

    if (!ALPHANUMERIC(c)) {
      separator1 = k;
    }

    if (c == '\"') {
      quotes1 = k;
    }
  }

  // Now scan the string from the last char back to the i-th char looking for the last
  // non alphanumeric char and the last quotes char
  for (k = StrLen(line)-1; k > i; k --) {
    c = line[k];

    if (!ALPHANUMERIC(c)) {
      separator2 = k;
    }

    if (c == '\"') {
      quotes2 = k;
    }
  }

  c = line[i];

  if (c != '\"' && quotes1 >= 0 && quotes2 > quotes1+1
   && quotes2-quotes1 < NCHARS) {
    // We have identified some text limited by quotes, send that text to the input field
    StrNCopy(string, line+quotes1+1, quotes2-quotes1-1);
    string[quotes2-quotes1-1] = ' ';
    string[quotes2-quotes1] = '\0';
    FldInsert(fld, string, StrLen(string));	// add the string after the cursor
    FrmSetFocus(frm, index);    		// set focus to the input field
  } else if (ALPHANUMERIC(c) && separator2 > separator1+1
          && separator2-separator1 < NCHARS) {
    // We have identified a word, send that word to the input field
    StrNCopy(string, line+separator1+1, separator2-separator1-1);
    string[separator2-separator1-1] = ' ';
    string[separator2-separator1] = '\0';
    FldInsert(fld, string, StrLen(string));	// add the string after the cursor
    FrmSetFocus(frm, index);    		// set focus to the input field
  }
}



/*
 * Change the handle of the string associated with a text field to the
 * specified handle, and free the previous handle from memory
 */
FieldPtr
SetFieldTextFromHandle(FormPtr frm, UInt16 fieldID, MemHandle txtH) {
  MemHandle oldTxtH;
  FieldPtr fldP;
  UInt16 objIdx;

  objIdx = FrmGetObjectIndex(frm, fieldID);
  fldP = FrmGetObjectPtr(frm, objIdx);
  oldTxtH = FldGetTextHandle(fldP);
  FldSetTextHandle(fldP, txtH);
  FldDrawField(fldP);
  if(oldTxtH) {
    MemHandleFree(oldTxtH);
  }
  return fldP;
}



/*
 * Set a text field to the specified string
 */
FieldPtr
SetFieldTextFromStr(FormPtr frm, UInt16 fieldID, char *strP) {
  MemHandle txtH;
  FieldPtr fldP;

  txtH = MemHandleNew(StrLen(strP) + 1);
  if(!txtH) {
    return NULL;
  }
  StrCopy(MemHandleLock(txtH), strP);
  fldP = SetFieldTextFromHandle(frm, fieldID, txtH);
  MemHandleUnlock(txtH);
  return fldP;
}



/*
 * Update the clocks
 */
void
UpdateClocks() {
  char TmWhSt[16], TmBlSt[16];
  int min, sec, hr;
  Boolean separator;

  int TmWh = MvTimeWh, TmBl = MvTimeBl;
  UInt32 CurrTm = TimGetSeconds();

  if (GetActiveFormId() != BoardForm) {
    return;
  }

  separator = true;

  if(MvTimeLast != -1 && Relation >= -1 && Relation <= 1) {
    // If we are playing or observing a game let the clock run
    if(BlToMv) {
      TmBl -= (CurrTm - MvTimeLast);
    } else {
      TmWh -= (CurrTm - MvTimeLast);
    }

    if (((TimGetTicks()-MvTickLast) / (SysTicksPerSecond()/3)) % 3 == 0) {
      // Each 1/3 of a second make the separator disappear from the running clock
      separator = false;
    }
  }

  if (TmWh < 0) {
    if (separator || BlToMv) {
      StrCopy(TmWhSt, "-:--");
    } else {
      StrCopy(TmWhSt, "- --");
    }
  } else {
    sec = TmWh%60;
    min = (TmWh/60)%60;
    hr = (TmWh/(60*60));

    if(hr) {
      if (separator || BlToMv) {
        StrPrintF(TmWhSt, "%d.%02d", hr, min);
      } else {
        StrPrintF(TmWhSt, "%d %02d", hr, min);
      }
    } else {
      if (separator || BlToMv) {
        StrPrintF(TmWhSt, "%d:%02d", min, sec);
      } else {
        StrPrintF(TmWhSt, "%d %02d", min, sec);
      }
    }
  }

  if (TmBl < 0) {
    if (separator || !BlToMv) {
      StrCopy(TmBlSt, "-:--");
    } else {
      StrCopy(TmBlSt, "- --");
    }
  } else {
    sec=TmBl%60;
    min=(TmBl/60)%60;
    hr =(TmBl/(60*60));

    if(hr) {
      if (separator || !BlToMv) {
        StrPrintF(TmBlSt, "%d.%02d", hr, min);
      } else {
        StrPrintF(TmBlSt, "%d %02d", hr, min);
      }
    } else {
      if (separator || !BlToMv) {
        StrPrintF(TmBlSt, "%d:%02d", min, sec);
      } else {
        StrPrintF(TmBlSt, "%d %02d", min, sec);
      }
    }
  }

  WinDrawChars(TmWhSt, StrLen(TmWhSt), 144, Flip ? 16 : 24);
  WinDrawChars(TmBlSt, StrLen(TmBlSt), 144, Flip ? 24 : 16);
}



/*
 * Update the chessboard
 */
void
UpdateBoard(Boolean newdata) {
  static char tmpstr[256];
  char *fields[48];
  int cll, nfields=0;
  int i, j;
  char *lastmove;

  if(newdata) {
    MvTimeLast = TimGetSeconds();
    MvTickLast = TimGetTicks();
    StrCopy(CurrentBoard, CurrentLine);

    if (NextMoveDestR != -1) {
      /* Clean markers */
      NextMoveStartR = NextMoveStartC = -1;
      NextMoveDestR = NextMoveDestC = -1;
    }
  }

  StrCopy(tmpstr, CurrentBoard);
  cll = StrLen(tmpstr);
  for(i=0; i<cll; i++) {
    if(tmpstr[i] == ' ') {
      fields[nfields++] = &tmpstr[i+1];
      tmpstr[i] = 0;
    }
  }

  GameNumber = StrAToI(fields[15]);
  WhiteName = fields[16];
  BlackName = fields[17];
  MvTimeWh = StrAToI(fields[23]);
  MvTimeBl = StrAToI(fields[24]);

  // If our name appears as the black player, and not as the white player,
  // we have to flip the Board
  Flip = (StrCompare(ChessPrefs.user, WhiteName) != 0
          && StrCompare(ChessPrefs.user, BlackName) == 0);

  BlToMv = (fields[8][0] == 'B');	// see whose turn it is

  if(newdata) {
    // Reset the auto-off timer whenever a new Board is received
    EvtResetAutoOffTimer();

    Relation = StrAToI(fields[18]);	// my relation with the game

    if (Relation == 1) {	// I'm playing and it's my turn
      // Wake me up!
      SndPlaySystemSound(sndInfo);

      // Never go off
      SysSetAutoOffTime(0);
    } else {	// Not my turn
      // Allow the device to go off as usual
      SysSetAutoOffTime(AutoOffTime);
    }

    /* Get the square of last move */
    lastmove = fields[28];

    if (StrCaselessCompare(lastmove, "O-O") == 0) {
      /* Make the king blink */
      LastMoveC = 6;
      LastMoveR = (BlToMv ? 0 : 7);
      LastMoveBlink = 1;
    } else if (StrCaselessCompare(lastmove, "O-O-O") == 0) {
      /* Make the king blink */
      LastMoveC = 2;
      LastMoveR = (BlToMv ? 0 : 7);
      LastMoveBlink = 1;
    } else {
      /* Read the destination square of last move */
      for (i = 0; lastmove[i] && lastmove[i+1]; i ++) {
        if (lastmove[i] >= 'a' && lastmove[i] <= 'h'
            && lastmove[i+1] >= '1' && lastmove[i+1] <='8') {
          /* Make such square blink */
          LastMoveC = lastmove[i] - 'a';
          LastMoveR = lastmove[i+1] - '1';
          LastMoveBlink = 1;

          break;
        }
      }
    }

    // If there is new board data but the Board form is not the main form,
    // switch to the Board form and return
    if(MainFormId != BoardForm) {
      FrmGotoForm(BoardForm);
      return;
    }
  }

  if (GetActiveFormId() == BoardForm) {
    UpdateClocks();

    for(i=0; i<8; i++) {
      for(j=0; j<8; j++) {
        Board[7-i][j] = fields[i][j];
        DrawSquare(7-i, j);
      } /* end for j */
    } /* end for i */
  }
} /* end UpdateBoard */



/*
 * Refresh the message area
 */
void
RefreshMessages() {
  int i, nrows, top, bottom, max;
  RectangleType oldClip, newClip;
  char (*buffer)[NCHARS+1];
  int formId;

  formId = GetActiveFormId();

  if (formId != BoardForm && formId != MessagesForm) {
    return;
  }

  // Calculate how many rows must be redrawn
  if(formId == BoardForm) {
    top = BOARDTOP + SquareSize*8;
    buffer = CurrentRows;
    bottom = 0;
    max = CurrentNRows;
  } else {
    top = MESSAGETOP;

    if (Current) {
      buffer = CurrentRows;
      bottom = 0;
      max = CurrentNRows;
    } else {
      buffer = SavedRows;
      bottom = SavedBottomRow;
      max = SavedNRows;
    }
  }

  nrows = (MESSAGEBOTTOM - top + FONTHEIGHT - 1) / FONTHEIGHT;

  // Set font and clipping region
  FntSetFont(TermFontID);

  WinGetClip(&oldClip);

  newClip.topLeft.x = 0;
  newClip.topLeft.y = top;
  newClip.extent.x = 160;
  newClip.extent.y = MESSAGEBOTTOM - top;

  WinClipRectangle(&newClip);
  WinSetClip(&newClip);

  // Draw from bottom to top
  for(i = 0; i < nrows; i ++) {
    if (bottom + i < max) {
      WinDrawChars(buffer[bottom + i], NCHARS, 0, MESSAGEBOTTOM - (i+1)*FONTHEIGHT);
    }
  }

  WinSetClip(&oldClip);
} /* end RefreshMessages */



/*
 * Load the bitmaps of the pieces into memory
 */
void
LoadBitmap(int size, int bmpno) {
  WinHandle tmpHandle, tmpHandle2;
  Err err;
  MemHandle bitmapHandle;
  int oldsize;
  RectangleType r;
  Coord width, height;

  oldsize = SquareSize;
  SquareSize = size;

  if(OffImage) {
    WinDeleteWindow(OffImage, false);
  }
  
  tmpHandle2 = WinGetDrawWindow();

  // Get the physical size of the screen
  WinGetDisplayExtent(&width, &height);

  // Allocate an off screen window of the right size even for hi density (very likely today)
  // Note: nativeFormat was defined in OS 5 but doesn't crash if used earlier
  OffImage = WinCreateOffscreenWindow((SquareSize*7*width)/160, (SquareSize*4*height)/160, nativeFormat, &err);

  tmpHandle = WinSetDrawWindow(OffImage);
  bitmapHandle = DmGetResource('Tbmp',bmpno);
  WinDrawBitmap(MemHandleLock(bitmapHandle),0,0);
  MemHandleUnlock(bitmapHandle);
  WinSetDrawWindow(tmpHandle2);

  if (GetActiveFormId() == BoardForm) {
    r.topLeft.x = 0;
    r.topLeft.y = 0;
    r.extent.x = oldsize * 8;
    r.extent.y = oldsize * 8;
    WinEraseRectangle(&r, 1);

    if (size > oldsize) {
      // Clear part of the message area below the current Board
      r.topLeft.x = 0;
      r.topLeft.y = oldsize * 8;
      r.extent.x = 160;
      r.extent.y = (size - oldsize) * 8;
      WinEraseRectangle(&r, 1);
    }

    UpdateBoard(false);
    RefreshMessages();
  }
}



/*
 * Handle the menus
 */
void
HandleMenu(EventPtr e) {
  FormPtr frm;
  UInt16 id;

  MenuEraseStatus(NULL);

  frm = FrmGetActiveForm();

  if (GetActiveFormId() == BoardForm) {
    id = BoardFld;
  } else {
    id = MessagesFld;
  }

  switch(e->data.menu.itemID) {
  case CmdMenuDraw:
    SetFieldTextFromStr(frm, id, "draw ");
    break;
  case CmdMenuResign:
    SetFieldTextFromStr(frm, id, "resign ");
    break;
  case CmdMenuSay:
    SetFieldTextFromStr(frm, id, "say ");
    break;
  case CmdMenuAccept:
    SetFieldTextFromStr(frm, id, "accept ");
    break;
  case CmdMenuDecline:
    SetFieldTextFromStr(frm, id, "decline ");
    break;
  case CmdMenuPlay:
    SetFieldTextFromStr(frm, id, "play ");
    break;
  case CmdMenuMatch:
    SetFieldTextFromStr(frm, id, "match ");
    break;
  case CmdMenuSeek:
    SetFieldTextFromStr(frm, id, "seek ");
    break;
  case CmdMenuUnseek:
    SetFieldTextFromStr(frm, id, "unseek ");
    break;
  case CmdMenuSought:
    SetFieldTextFromStr(frm, id, "sought ");
    break;
  case CmdMenuTell:
    SetFieldTextFromStr(frm, id, "tell ");
    break;
  case CmdMenuHelp:
    SetFieldTextFromStr(frm, id, "help ");
    break;
  case CmdMenuQuit:
    SetFieldTextFromStr(frm, id, "quit ");
    break;
  case BoardMenu16:
    LoadBitmap(16, PieceBmp16);
    if (MainFormId != BoardForm) {
      FrmGotoForm(BoardForm);
    }
    break;
  case BoardMenu18:
    LoadBitmap(18, PieceBmp18);
    if (MainFormId != BoardForm) {
      FrmGotoForm(BoardForm);
    }
    break;
  case OptMenuAbout:
    FrmPopupForm(AboutForm);
    break;
  case OptMenuHelp:
    FrmHelp(OptHelpStr);
    break;
  }
}



/*
 * Handle events of the login form
 */
Boolean
LoginFormHandleEvent(EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  char *serverPtr, *portPtr, *userPtr;
  NetSocketAddrINType address;
  NetIPAddr addr;
  Int32 port;
  NetHostInfoBufType buffer;
  NetHostInfoType *host;
  Err err;
  Int16 i;
  UInt16 index;
  ControlPtr ctlPassword;
    
  frm = FrmGetActiveForm();
  index = FrmGetObjectIndex(frm, PassSel);

  switch (e->eType) {
  case frmOpenEvent:
    FrmDrawForm(frm);
    MainFormId = FrmGetFormId(frm);

    LoadPrefs();

    SockConnected = false;	// probably not needed, but safe

    SetFieldTextFromStr(frm, ServerFld, ChessPrefs.host);
    SetFieldTextFromStr(frm, PortFld, ChessPrefs.port);
    SetFieldTextFromStr(frm, UserFld, ChessPrefs.user);

    ctlPassword = FrmGetObjectPtr(frm, index);
    CtlSetLabel(ctlPassword, StrLen(ChessPrefs.pass) == 0 ? "-Empty-" : "-Assigned-");

    handled = true;
    break;

  case frmUpdateEvent:
    if(e->data.frmUpdate.updateCode == MYUPDATECODE) {
      // The update was requested by the password form.
      // Only the password select trigger needs updating.
      ctlPassword = FrmGetObjectPtr(frm, index);
      CtlSetLabel(ctlPassword, StrLen(ChessPrefs.pass) == 0 ? "-Empty-" : "-Assigned-");
      handled = true;
    }
    break;

  case menuEvent:
    HandleMenu(e);
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case PassSel:
      FrmPopupForm(PasswordForm);
      break;

    case ConnectBtn:
      if ((err = InitNetwork())) {
        switch(err) {
        case netErrOutOfMemory:
          PopupErrorAlert(OutMemAlert, err);
          break;
        case netErrUserCancel:
          PopupErrorAlert(UserCancelAlert, err);
          break;
        default:
          PopupErrorAlert(NetLibAlert, err);
          break;
        }

        break;
      }

      serverPtr = FldGetTextPtr(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ServerFld)));
      portPtr   = FldGetTextPtr(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PortFld)));
      userPtr   = FldGetTextPtr(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, UserFld)));

      if (!serverPtr) { serverPtr = ""; }
      if (!portPtr)   { portPtr = ""; }
      if (!userPtr)   { userPtr = ""; }

      StrCopy(ChessPrefs.host, serverPtr);
      StrCopy(ChessPrefs.port, portPtr);
      StrCopy(ChessPrefs.user, userPtr);

      // Convert the port field into a 16 bit integer
      port = StrAToI(ChessPrefs.port);

      if (port <= 0 || port >= 65536) {
        (void) FrmAlert(PortAlert);
        break;
      }

      // Convert the server field into a NetIPAddr address
      // First try it as aaa.bbb.ccc.ddd
      addr = NetLibAddrAToIN(NetLibRefnum, ChessPrefs.host);

      if (addr == -1) {
        // Now try it as a host name. Display a message while this happens
        PopupNowForm(DNSWaitForm);
        host = NetLibGetHostByName(NetLibRefnum, ChessPrefs.host, &buffer,
                                 SysTicksPerSecond()*3, &err);
        FrmReturnToForm(0);

        if (host != NULL) {
          addr = buffer.address[0];
        } else {
          switch (err) {
          case netErrTimeout:
          case netErrDNSTimeout:
            PopupErrorAlert(DNSTimeoutAlert, err);
            break;
          case netErrDNSNameTooLong:
          case netErrDNSBadName:
          case netErrDNSLabelTooLong:
          case netErrDNSNonexistantName:
            PopupErrorAlert(ServerNameAlert, err);
            break;
          case netErrNoDNSServers:
            PopupErrorAlert(NoDNSAlert, err);
            break;
          default:
            PopupErrorAlert(DNSAlert, err);
            break;
          }

          break;
        }
      }

      address.family = netSocketAddrINET;
      address.port = port;
      address.addr = addr;

      // Try to get a socket within 1 second
      Sock = NetLibSocketOpen(NetLibRefnum, netSocketAddrINET, netSocketTypeStream,
                          netSocketProtoIPTCP, SysTicksPerSecond(), &err);

      if (Sock == -1) {
        if (err == netErrOutOfMemory) {
          PopupErrorAlert(OutMemAlert, err);
        } else {
          PopupErrorAlert(SocketAlert, err);
        }

        break;
      }

      // If we are here then the socket was created successfully
      // therefore we have only to connect it to the server.
      // Connect the socket within 3 secs. Display a message while this happens
      PopupNowForm(ConnectWaitForm);
      i = NetLibSocketConnect(NetLibRefnum, Sock, (NetSocketAddrType *) &address,
                          sizeof(address), SysTicksPerSecond()*3, &err);
      FrmReturnToForm(0);

      if (i != 0 && err != 0) {
        if (err == netErrTimeout) {
          PopupErrorAlert(ServerTimeoutAlert, err);
        } else {
          PopupErrorAlert(ServerAlert, err);
        }

        NetLibSocketClose(NetLibRefnum, Sock, SysTicksPerSecond(), &err);
        break;
      }

      SockConnected = true;

      StrCopy(CurrentLine, "");
      StrPrintF(InitStr, "%s\n%s\nset style 12\n", ChessPrefs.user, ChessPrefs.pass);
      SentInitStr = false;

      SavePrefs();

      FrmGotoForm(MessagesForm);
    }

    handled = true;
    break;

  default:
    break;
  }

  return handled;
}



/*
 * Handle events of the password form (the one popped up by
 * the password select trigger)
 */
Boolean
PasswordFormHandleEvent(EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  UInt16 index;
  FieldPtr fld;
  char *password;
    
  frm = FrmGetActiveForm ();
  index = FrmGetObjectIndex(frm, PassFld);

  switch (e->eType) {
  case frmOpenEvent:
    FrmDrawForm(frm);

    // Clean the password field
    SetFieldTextFromStr(frm, PassFld, "");

    // Set focus to password field
    FrmSetFocus(frm, index);

    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case PassCancelBtn:
      // Just return to login form
      FrmReturnToForm(0);
      handled = true;
      break;

    case PassOkBtn:
      fld = FrmGetObjectPtr(frm, index);
      password = FldGetTextPtr(fld);

      if (password == NULL) {
        password = "";
      }

      // Copy the password to global field
      StrCopy(ChessPrefs.pass, password);

      // When this form closes update the login form so that
      // the password select trigger gets updated
      FrmUpdateForm(LoginForm, MYUPDATECODE);

      FrmReturnToForm(0);
      handled = true;
      break;

    default:
      break;
    }
    break;

  default:
    break;
  }

  return handled;
}



/*
 * Send the current string contained in the field identified by id to the server
 */
void
SendCommand(FormPtr frm, UInt16 id) {
  FieldPtr fld;
  char *command;
  static char tmpstr[256];
  Err err;
  UInt16 index;

  index = FrmGetObjectIndex(frm, id);
  fld = FrmGetObjectPtr(frm, index);
  command = FldGetTextPtr(fld);

  if (command == NULL) {
    command = "";
  }

  // Send the command on screen
  PutStr(command);

  // Send the command to the server
  if (StrLen(command) < 254 && SockConnected) {
    StrCopy(tmpstr, command);
    StrCat(tmpstr, "\n");
    NetLibSend(NetLibRefnum, Sock, (void *)tmpstr, StrLen(tmpstr),
     0, NULL, 0, SysTicksPerSecond()*5, &err);
  }

  // Clean the field
  SetFieldTextFromStr(frm, id, "");
}



/*
 * Scroll up messages. Scroll buttons are updated accordingly
 */
void
ScrollUp(FormPtr frm, UInt16 upIndex, UInt16 dnIndex) {
  int nrows;

  nrows = (MESSAGEBOTTOM - MESSAGETOP) / FONTHEIGHT;
  SavedBottomRow += SCROLLAMOUNT;

  if (SavedBottomRow+nrows >= SavedNRows) {
    SavedBottomRow = SavedNRows-nrows;
    if (SavedBottomRow < 0) {
      SavedBottomRow = 0;
    }
  }

  RefreshMessages();

  FrmUpdateScrollers(frm, upIndex, dnIndex,
   SavedBottomRow + (MESSAGEBOTTOM-MESSAGETOP)/FONTHEIGHT < SavedNRows, SavedBottomRow > 0);
}



/*
 * Scroll down messages. Scroll buttons are updated accordingly
 */
void
ScrollDn(FormPtr frm, UInt16 upIndex, UInt16 dnIndex) {
  SavedBottomRow -= SCROLLAMOUNT;

  if (SavedBottomRow <= 0) {
    SavedBottomRow = 0;
  }

  RefreshMessages();

  FrmUpdateScrollers(frm, upIndex, dnIndex,
   SavedBottomRow + (MESSAGEBOTTOM-MESSAGETOP)/FONTHEIGHT < SavedNRows, SavedBottomRow > 0);

}



/*
 * Handle events of the message form
 */
Boolean
MessagesFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  UInt16 index, upIndex, dnIndex;
  int i;
  int y;
  ControlPtr ctlCurrent, ctlSaved;
  FieldPtr fld;
    
  frm = FrmGetActiveForm();
  upIndex = FrmGetObjectIndex(frm, ScrollUpBtn);
  dnIndex = FrmGetObjectIndex(frm, ScrollDnBtn);

  switch (e->eType) {
  case frmOpenEvent:
  case frmUpdateEvent:
    FrmDrawForm(frm);

    MainFormId = FrmGetFormId(frm);

    index = FrmGetObjectIndex(frm, CurrBtn);
    ctlCurrent = FrmGetObjectPtr(frm, index);
    index = FrmGetObjectIndex(frm, SavedBtn);
    ctlSaved = FrmGetObjectPtr(frm, index);

    if (Current) {
      CtlSetValue(ctlCurrent, 1);
      CtlSetValue(ctlSaved, 0);
      FrmUpdateScrollers(frm, upIndex, dnIndex, false, false);
    } else {
      CtlSetValue(ctlCurrent, 0);
      CtlSetValue(ctlSaved, 1);
      FrmUpdateScrollers(frm, upIndex, dnIndex,
       SavedBottomRow + (MESSAGEBOTTOM-MESSAGETOP)/FONTHEIGHT < SavedNRows, SavedBottomRow > 0);
    }

    RefreshMessages();

    if(e->eType == frmOpenEvent) {
      index = FrmGetObjectIndex(frm, MessagesFld);
      fld = FrmGetObjectPtr(frm, index);
      FldSetInsPtPosition(fld, FldGetTextLength(fld));	// Put the cursor at the end
      FrmSetFocus(frm, index);	// Set focus to the input field
    }

    handled = true;
    break;

  case winEnterEvent:
    if (e->data.winEnter.enterWindow == FrmGetWindowHandle(frm)) {
      // Entering the form's window that was previously covered
      RefreshMessages();
    }
    break;

  case menuEvent:
    HandleMenu(e);
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case ToBoardBtn:
      FrmGotoForm(BoardForm);
      handled = true;
      break;
    case CurrBtn:
      if (!Current) {
        FrmUpdateScrollers(frm, upIndex, dnIndex, false, false);
        Current = true;
        RefreshMessages();
      }
      handled = true;
      break;
    case SavedBtn:
      if (Current) {
        // Copy the current buffer to the saved one
        SavedNRows = CurrentNRows;
        SavedBottomRow = 0;

        for (i = 0; i < CurrentNRows; i ++) {
          StrCopy(SavedRows[i], CurrentRows[i]);
        }

        FrmUpdateScrollers(frm, upIndex, dnIndex, (MESSAGEBOTTOM-MESSAGETOP)/FONTHEIGHT < SavedNRows, false);

        Current = false;
        RefreshMessages();
      }
      handled = true;
      break;
    case MessagesClearBtn:
      // Clear button pressed. Clean the input field
      SetFieldTextFromStr(frm, MessagesFld, "");
      handled = true;
      break;
    case MessagesSendBtn:
      // Send button pressed. Send the command to the server
      SendCommand(frm, MessagesFld);

      if (!Current) {
        Current = true;
	// Update radio buttons and messages
        FrmUpdateForm(MessagesForm, MYUPDATECODE);
      }

      handled = true;
      break;
    default:
      break;
    }
    break;

  case ctlRepeatEvent:
    if (!Current) {
      switch(e->data.ctlRepeat.controlID) {
      case ScrollUpBtn:
        ScrollUp(frm, upIndex, dnIndex);
        break;
      case ScrollDnBtn:
        ScrollDn(frm, upIndex, dnIndex);
        break;
      default:
        break;
      }

      // DO NOT set handled to true, because the ctlRepeatEvent must go
      // back to the system in order to generate another ctlRepeatEvent
      // if the button stays pressed
    }
    break;

  case keyDownEvent:
    switch(e->data.keyDown.chr) {
    case '\n':
      // Send the command to the server
      SendCommand(frm, MessagesFld);

      if (!Current) {
        Current = true;
	// Send an update event to the form so as to update radio buttons and messages
        FrmUpdateForm(MessagesForm, MYUPDATECODE);
      }

      handled = true;
      break;
    default:
      break;
    }
    break;

  case penDownEvent:
    y = e->screenY;

    if (y >= MESSAGETOP && y < MESSAGEBOTTOM) {
      HandlePenOnMessage(e->screenX, y);
      handled = true;
    } 
    break;

  default:
    break;
  }

  return handled;
}



/*
 * Handle events of the board form
 */
Boolean
BoardFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
  int y;
  UInt16 index;
  FieldPtr fld;

  frm = FrmGetActiveForm();
    
  switch (e->eType) {
  case frmOpenEvent:
  case frmUpdateEvent:
    FrmDrawForm(frm);

    MainFormId = FrmGetFormId(frm);
    UpdateBoard(false);
    RefreshMessages();

    if(e->eType == frmOpenEvent) {
      index = FrmGetObjectIndex(frm, BoardFld);
      fld = FrmGetObjectPtr(frm, index);
      FldSetInsPtPosition(fld, FldGetTextLength(fld));	// Put the cursor at the end
      FrmSetFocus(frm, index);	// Set focus to the input field
    }

    handled = true;
    break;

  case winEnterEvent:
    if (e->data.winEnter.enterWindow == FrmGetWindowHandle(frm)) {
      // Entering the form's window that was previously covered
      UpdateBoard(false);
      RefreshMessages();
    }
    break;

  case menuEvent:
    HandleMenu(e);
    handled = true;
    break;

  case penDownEvent:
    y = e->screenY;

    if (y >= BOARDTOP) {
      if (y < BOARDTOP + SquareSize*8) {
        if (e->screenX < SquareSize*8) {
          HandlePenOnBoard(e->screenX, y);
          handled = true;
        }
      } else if (y < MESSAGEBOTTOM) {
        HandlePenOnMessage(e->screenX, y);
        handled = true;
      }
    } 
    break;

  case keyDownEvent:
    if (e->data.keyDown.chr == '\n') {
      // Send the command to the server
      SendCommand(frm, BoardFld);
      handled = true;
    }
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case ToMessagesBtn:
      FrmGotoForm(MessagesForm);
      handled = true;
      break;
    case BoardClearBtn:
      // Clear button pressed. Clean the input field
      SetFieldTextFromStr(frm, BoardFld, "");
      handled = true;
      break;
    case BoardSendBtn:
      // Send button pressed. Send the command to the server
      SendCommand(frm, BoardFld);
      handled = true;
      break;
    default:
      break;
    }
    break;

  default:
    break;
  }

  return handled;
}



/*
 * Handle events of the about form
 */
Boolean
AboutFormHandleEvent (EventPtr e) {
  Boolean handled = false;
  FormPtr frm;
    
  switch (e->eType) {
  case frmOpenEvent:
    frm = FrmGetActiveForm();
    FrmDrawForm(frm);
    handled = true;
    break;
  case ctlSelectEvent:
    if (e->data.ctlSelect.controlID == FormAboutButton) {
      // When the about form closes, update the main form so that
      // messages come in the meanwhile can be displayed
/*      if (MainFormId == BoardForm) {
        FrmUpdateForm(BoardForm, MYUPDATECODE);
      } else if (MainFormId == MessagesForm) {
        FrmUpdateForm(MessagesForm, MYUPDATECODE);
      }
*/
      FrmReturnToForm(0);
      handled = true; 
      break;
    }
    break;
  default:
    break;
  }

  return handled;
}



/*
 * Process a line of characters received from the server.
 * The line must be stored in CurrentLine prior to calling this function.
 */
Boolean
ApplicationHandleLine() {
  int n;
  static char tmpstr[256];
  char *ptr;
  
  if (StrNCompare("<12> ", CurrentLine, 5) == 0) {
    // Received a Board
    UpdateBoard(true);
    return true;
  }

  if (Relation >= -1 && Relation <= 1) {
    // While we are playing or observing a game, look if it's over
    ptr = StrStr(CurrentLine, "{Game ");

    if (ptr != NULL) {
      ptr += 6;	// skip the word "{Game "

      while (*ptr == ' ') ptr ++;	// skip blanks

      if (*ptr >= '0' && *ptr <= '9') {
        // Calculate the number after the word "{Game "
        n = 0;
        while (*ptr >= '0' && *ptr <= '9') {
          n = n*10 + (*ptr) - '0';
          ptr ++;
        }

        if (*ptr != '\0') ptr++;		// skip next character

        while (*ptr == ' ') ptr++;	// skip blanks

        if (n == GameNumber) {
          // There has been a notification exactly about the game we
          // are playing or observing, so assume it's over
          Relation = -3;

          // Pop up a window to inform about how it's ended
          StrCopy(tmpstr, ptr);
          ptr = StrChr(tmpstr, '}');
          if (ptr != NULL) {
            *ptr = ' ';	// remove eventual closing brace
          }

          (void) FrmCustomAlert(EndGameAlert, tmpstr, " ", " ");
        }
      }
    }
  }

  PutStr(CurrentLine);

  return true;
}



/*
 * Wait for the server for half a second and if there is new data process it.
 * Return false in case of an error.
 */
Boolean
ApplicationHandleData() {
  Err err;
  Int16 i;
  int cll, j;
  static char receiveBuffer[256];

  cll = StrLen(CurrentLine);

  if(!SentInitStr) {
    PutStr("ChessClient% Connecting to the server...");
  }

  i = NetLibReceive(NetLibRefnum, Sock, receiveBuffer, sizeof(receiveBuffer),
                    0, NULL, NULL, SysTicksPerSecond()/2, &err);

  if(i == 0) {
    // The server closed connection. Scroll up 10 rows so that last
    // messages can be read even after the alert pops up
    for (j = 0; j < 10; j ++) PutStr("");

    // Close network
    NetLibSocketClose(NetLibRefnum, Sock, SysTicksPerSecond(), &err);
    NetLibClose(NetLibRefnum, false);
    Sock = -1;
    SockConnected = false;

    // Pop up an alert then return to login form
    FrmAlert(QuitAlert);
    FrmGotoForm(LoginForm);
    return false;
  } else if(i == -1) {
    StrPrintF(receiveBuffer, "ChessClient% Network error (error number: 0x%04x)", err);
    PutStr(receiveBuffer);
    return false;
  }

  if(!SentInitStr) {
    PutStr("ChessClient% Ok");
    PutStr("ChessClient% Sending user/password...");

    if (SockConnected) {
      NetLibSend(NetLibRefnum, Sock, (void *)InitStr, StrLen(InitStr),
       0, NULL, 0, SysTicksPerSecond()*5, &err);
    }

    PutStr("ChessClient% Ok");
    SentInitStr = true;
  }

  for(j=0; j<i; j++) {
    if(receiveBuffer[j] == '\n' || receiveBuffer[j] == '\r' || cll == 255) {
      if(cll) {
        ApplicationHandleLine();
        StrCopy(CurrentLine, "");
        cll = 0;
      }
      continue;
    }

    if(receiveBuffer[j] >= ' ') {
      // Add any printable char to CurrentLine
      CurrentLine[cll++] = receiveBuffer[j];
      CurrentLine[cll] = 0;
    }
  }

  return true;
}
  

/*
 * Process events before SysHandleEvent(). This allows special behaviour
 * for some system event. In our case we need to process scrolling hard keys
 * without getting them "digested" as system events.
 */
Boolean
PreprocessEvent(EventPtr e) {
  FormPtr frm;
  UInt16 upIndex, dnIndex;
  Boolean handled = false;
  static char command[10];

  if (GetActiveFormId() == MessagesForm && e->eType == keyDownEvent && !Current) {
    frm = FrmGetActiveForm();
    upIndex = FrmGetObjectIndex(frm, ScrollUpBtn);
    dnIndex = FrmGetObjectIndex(frm, ScrollDnBtn);

    switch(e->data.keyDown.chr) {
    case vchrPageUp:
    case vchrRockerUp:
    case vchrThumbWheelUp:
    case chrUpArrow:
      ScrollUp(frm, upIndex, dnIndex);
      handled = true;
      break;
    case vchrPageDown:
    case vchrRockerDown:
    case vchrThumbWheelDown:
    case chrDownArrow:
      ScrollDn(frm, upIndex, dnIndex);
      handled = true;
      break;
    default:
      break;
    }
  }

  return handled;
}


/*
 * Function that processes frmLoadEvent events by setting
 * the appropriate event handler to the form being loaded
 */
Boolean
ApplicationHandleEvent(EventPtr e) {
  FormPtr frm;
  UInt16 formId;
  Boolean handled = false;

  if (e->eType == frmLoadEvent) {
    formId = e->data.frmLoad.formID;
    frm = FrmInitForm(formId);
    FrmSetActiveForm(frm);
    switch(formId) {
    case BoardForm:
      FrmSetEventHandler(frm, BoardFormHandleEvent);
      break;
    case MessagesForm:
      FrmSetEventHandler(frm, MessagesFormHandleEvent);
      break;
    case LoginForm:
      FrmSetEventHandler(frm, LoginFormHandleEvent);
      break;
    case PasswordForm:
      FrmSetEventHandler(frm, PasswordFormHandleEvent);
      break;
    case AboutForm:
      FrmSetEventHandler(frm, AboutFormHandleEvent);
      break;
    }
    handled = true;
  }

  return handled;
}



/*
 * Function invoked when the application starts.
 * Get auto-off time, load fonts and bitmaps
 */
UInt16
StartApplication(void) {
  MemHandle fontHandle;
  FontType *fontPtr;
  int i;

  // Read the auto-off time without changing it
  AutoOffTime=SysSetAutoOffTime(0);
  SysSetAutoOffTime(AutoOffTime);

  LoadBitmap(16, PieceBmp16);

  for (i = 0; i < NROWS; i++) {
    SavedRows[i][0] = '\0';
  }

  // 'NFNT' and TermFont are resource type and ID, defined by you in
  // resource description
  fontHandle = DmGetResource('nfnt', TermFont);
  fontPtr = MemHandleLock(fontHandle);

  // User defined fonts start from 128
  FntDefineFont(TermFontID, fontPtr);

  return 0;
}



/*
 * Function invoked when the application stops.
 * Save preferences and close forms
 */
void
StopApplication(void) {
  Err err;

  // Restore the initial auto-off time in case it's changed
  SysSetAutoOffTime(AutoOffTime);

  // Send a frmSave event to all the open forms.
  FrmSaveAllForms ();

  // Close all the open forms.
  FrmCloseAllForms ();

  // Unlock the piece bitmaps
  if(OffImage) {
    WinDeleteWindow(OffImage, false);
  }

  if(Sock != -1) {
    NetLibSocketClose(NetLibRefnum, Sock, SysTicksPerSecond(), &err);
  }

  NetLibClose(NetLibRefnum, false);
  SockConnected = false;

  SavePrefs();
}



/*
 * The main event loop
 */
void
EventLoop(void) {
  UInt16 err;
  EventType e;
  NetFDSetType readFDs, writeFDs, exceptFDs, width;

  StrCopy(CurrentBoard,
     "<12> rnbqkbnr pppppppp -------- -------- -------- -------- "
     "PPPPPPPP RNBQKBNR W -1 1 1 1 1 0 0 White Black -3 2 12 39 39 "
     "-1 -1 1 none (0:00) none 0");
  
  for(;;) {
    if (SockConnected) {
      // If socket is connected, check for new data from the server
      netFDZero(&readFDs);
      netFDSet(Sock, &readFDs);
      width = Sock;

      NetLibSelect(NetLibRefnum, width+1, &readFDs, &writeFDs, &exceptFDs,
		   SysTicksPerSecond()/20, &err);
      if (netFDIsSet(Sock, &readFDs)) {
        ApplicationHandleData();
      }
    }

    EvtGetEvent(&e, SysTicksPerSecond()/20);

    if (e.eType != nilEvent) {
      if (PreprocessEvent(&e)) {
        continue;
      }
      if (SysHandleEvent(&e)) {
        continue;
      }
      if (MenuHandleEvent(NULL, &e, &err)) {
        continue;
      }
      if (ApplicationHandleEvent(&e)) {
        continue;
      }

      FrmDispatchEvent (&e);
    
      if(e.eType == appStopEvent) {
        // If we are still connected display an alert before
        // allowing the application to exit
        if (!SockConnected || FrmAlert(AppStopAlert) == 0) {
          return;
        }
      }
    }

    UpdateClocks();
    if (LastMoveBlink) {
      DrawSquare(LastMoveR, LastMoveC);
    }
  }
}



/*
 * The "main" function in Palm OS
 */
UInt32
PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags) {
  UInt16 err;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    err = StartApplication();

    if (err) {
      return err;
    }

    MainFormId = -1;
    FrmGotoForm(LoginForm);

    EventLoop();
    StopApplication();
  }

  return 0;
}
