/*
 * @(#)bitmap2font.c
 *
 * Derived from: bitmap2c Copyright 2000, Aaron Ardiri (mailto:aaron@ardiri.com).
 * Modified by Marco Brenco with author's permission on July 2008
 * to make it generate font text files readable by PilRC
 * for the ChessClient project. It assumes that bitmaps are composed by 128 glyphs exactly.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

/**
 * The main-line.
 *
 * @param argc the number of command line arguments.
 * @param argv an pointer to a series of command line argument strings.
 * @return 0 on success, 1 otherwise.
 */
int
main(int argc, char *argv[])
{
  int result = 0;

  // called correctly?
  if (argc > 1) {
    unsigned short i;
    FILE     *outFile = NULL;

    // open the "output" file
    outFile = fopen("font.txt","w");

    // process all data files passed in
    for (i=1; i<argc; i++) {
      FILE           *inFile       = NULL;
      unsigned char  buffer[32768] = { 0 }; // 32K of memory
      unsigned char  outbuf[32768] = { 0 }; // 32K of memory
      unsigned short size, outsize, width, height, j, k;
      unsigned short charBytes;
      unsigned char data;

      // open the PilRC resource image file
      inFile = fopen(argv[i],"rb");
     
      // does it exist?
      if (inFile != NULL) {
        // lets skip over the "16 byte" header for PilRC output
        fseek(inFile, 0L, SEEK_SET); 

        // read in compress the "bitmap"
        size    = fread(buffer,1,16,inFile);
        outsize = fread(outbuf,1,32768,inFile);
        width  = (unsigned short)(((int)buffer[0] << 8) | (int)buffer[1]);
        height = (unsigned short)(((int)buffer[2] << 8) | (int)buffer[3]);

        // The width of any character, in bytes (every byte is 2 pixels)
        charBytes = width / 256;

        // Print PilRC header
        fprintf(outFile, "// A %dx%d fixed width font\n", charBytes*2, height);
        fprintf(outFile, "\n");
        fprintf(outFile, "fontType 36864\n");
        fprintf(outFile, "ascent ?\n");
        fprintf(outFile, "descent ?\n");

        for (i = 0; i < 128; i ++) {
          if (i == 9 || (i >= 32 && i <= 126)) {
            fprintf(outFile, "\n");
            fprintf(outFile, "GLYPH ");

            if (i == 9 || i == 32 || i == 39 || i == 92) {
              fprintf(outFile, "%d\n", i);
            } else {
              fprintf(outFile, "'%c'\n", i);
            }

            for (j = 0; j < height; j ++) {
              for (k = 0; k < charBytes; k ++) {
                data = outbuf[j*(128*charBytes) + i*charBytes + k];

                if (data & 0xf0) {
                  fprintf(outFile, "#");
                } else {
                  fprintf(outFile, "-");
                }

                if (data & 0x0f) {
                  fprintf(outFile, "#");
                } else {
                  fprintf(outFile, "-");
                }
              }	// for k

              fprintf(outFile, "\n");
            }	// for j
          }	// if
        }	// for i

        // close input file
        fclose(inFile);
      } else {
        printf("ERROR: [%s] FILE NOT FOUND\n", argv[i]);
      }
    }

    printf("\n");

    // close output file
    fclose(outFile);
  } else {
    printf("USAGE:\n");
    printf("  bitmap2font { file1.bin file2.bin ... fileN.bin }\n\n");

    result = 1;
  }

  return result;
}
